/*
 * Help Access Library 
 * A Library to access the contents of Windows Help files.
 *
 * Copyright (C) 1995-2000 Bernd Herd, http://www.herdsoft.com
 *                                                                  
 * This program is free software; you can redistribute it and/or    
 * modify it under the terms of the GNU General Public License as   
 * published by the Free Software Foundation; either version 2 of   
 * the License.              
 *                                                                  
 * This program is distributed in the hope that it will be useful,  
 * but WITHOUT ANY WARRANTY; without even the implied warranty of   
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    
 * GNU General Public License for more details.                     
 *                                                                  
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software      
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.        
 */
/* ENUMKW.C -------------------------------------------
 * @doc
 * @module ENUMKW.C |
 * Enumerates the list of Keywords contained in a Help file.
 *
 * Help Access Library Project.
 *
 *-----------------------------------------------------*/

#include <windows.h>
#include <windowsx.h>
#pragma argsused
#include <time.h>
#include <stdio.h>
#include <string.h>
#include "hlpacces.h"


BOOL LOCAL nEnumerateKeywords(HIFSFILE hKWBTree, KWENUMPROC fnKWEnumProc, LPCSTR StartKeyWord, LPARAM lParam);
BOOL LOCAL nEnumeratePage(HIFSFILE hKWBTree, int PageCount, int CurrentLevel, int PageSize, int CurrPage, KWENUMPROC fnKWEnumProc, LPCSTR StartKeyWord, LPARAM lParam);


/* @func BOOL | HlpEnumKeyWords |
 *
 * This Function enumerates the List of Keywords from a Help or Media Viewer
 * Tile. This Function is relatively fast.
 *
 * Note that one Help or Media Viewer file can Have several Keyword lists.
 * Usually in Help Files the Keyword list is named "K" in Media Viewe it's
 * Named "0K", the ALink-List in Windows 95 Files is named "A", others are
 * also possible.
 *
 * @rdesc
 * TRUE if the Function was Successfull
 * FALSE else
 */
BOOL WINAPI DLLEXP HlpEnumKeyWords(
                HIFS ifs,                       // @parm Handle to a opened File-System opened by <f IFSOpen>
                KWENUMPROC fnKWEnumProc,        // @parm <t KWENUMPROC> Application-Defined Callback-Procedure to be called for every Keyword retrieved. from the help file
                LPCSTR StartKeyWord,            // @parm Keyword to start the scan with.
                LPCSTR KWListName,              // @parm Name of the Keyword Table (K,0K,A...)
                LPARAM lParam)                  // @parm Aplication-defined parameter to be transfered to the Callback-function
{ BOOL      Success=FALSE;
  HIFSFILE  hKWBTree = IFSOpenFile(ifs, KWListName);

  if (hKWBTree)
  {
      Success = nEnumerateKeywords(hKWBTree, fnKWEnumProc, StartKeyWord, lParam);

    IFSCloseFile(hKWBTree);
  }
  return Success;
}


/* nEnumerateKeywords
 *
 * Does the bunch of Work to do enumeration
 */
BOOL LOCAL nEnumerateKeywords(HIFSFILE hKWBTree, KWENUMPROC fnKWEnumProc, LPCSTR StartKeyWord, LPARAM lParam)
{ BOOL            Success;

  BTREEHEADER     BTreeHdr;
  DWORD           dwSize;
  UINT            PageCount;


  //--------- Read Title B-Tree header ------------------
  IFSReadFile(hKWBTree, &BTreeHdr, sizeof(BTreeHdr));
  dwSize = IFSSeekFile(hKWBTree, 0L, 2);

  PageCount = (UINT) (dwSize/leword(BTreeHdr.PageSize));

  Success = nEnumeratePage(hKWBTree, PageCount, leword(BTreeHdr.NLevels)-1, leword(BTreeHdr.PageSize), leword(BTreeHdr.RootPage), fnKWEnumProc, StartKeyWord, lParam);

  return Success;
}

#pragma pack(1)
typedef struct
        { LEWORD  Count;
          LEDWORD DataOffset;
        } PACKED KWREC, NEAR *NPKWREC;
#pragma pack()


BOOL LOCAL nEnumeratePage(HIFSFILE hKWBTree, int PageCount, int CurrentLevel, int PageSize, int CurrPage, KWENUMPROC fnKWEnumProc, LPCSTR StartKeyWord, LPARAM lParam)
{ BOOL                  Success    = TRUE;
  DWORD                 loc        = sizeof(BTREEHEADER) + PageSize * (DWORD) CurrPage;
  NPSTR                 page       = (NPSTR) LocalAlloc(LPTR, PageSize),
                        nxt;
  NPKWREC               npKWRec;
  BTREENODEHEADER NEAR *CurrNode    = (BTREENODEHEADER NEAR *)page;       /* Current Node in B-Tree  */
  int                   i;

  if (page)
  {
     //----------- Read Page ----------------------------------------
     IFSSeekFile(hKWBTree, loc, 0);
     if (PageSize == IFSReadFile(hKWBTree, page, PageSize))
     { 
      if (CurrentLevel > 0)
      {  Success = TRUE;
         //-- We're in a node, search for next sub-node or leaf --
         for (i=0,                             nxt = page+sizeof(BTREEINDEXHEADER);
              i<leword(CurrNode->NEntries)+1 && Success;
              ++i,                             nxt+=lstrlen(nxt+2)+1+sizeof(WORD))
         {

             //------- We're currently in a node, so recurse down... ----
             if (lstrcmpi(nxt+2,StartKeyWord)>=0 || i==leword(CurrNode->NEntries))
                Success &= nEnumeratePage(hKWBTree, PageCount, CurrentLevel-1, PageSize, CurrPage=leword(((LEWORD NEAR *) (nxt))[0]), fnKWEnumProc, StartKeyWord, lParam);
         }
      }
      else
      {
         //-- We're in the right leaf now, scan for Data ---------
         for (i=0,                  nxt = page+sizeof(BTREENODEHEADER);
              i<leword(CurrNode->NEntries) && Success;
              ++i                  ,nxt+=lstrlen(nxt)+1+sizeof(KWREC))
            {npKWRec=(NPKWREC)(nxt+lstrlen(nxt)+1);
             if (lstrcmpi(nxt,StartKeyWord)>=0)
                Success &= fnKWEnumProc(nxt, ledword(npKWRec->DataOffset), leword(npKWRec->Count), lParam);
            }
      }
/*       //--------- Are we in a Node or in a leaf? -------------------
       if (CurrentLevel > 0)
       { //-- We're in a node, search for next sub-node or leaf --
         for (i=0,                  CurrIndexEntry = (BTREEINDEXENTRY NEAR *) (page+sizeof(BTREEINDEXHEADER));
              i<NodeHeader.NEntries+1;
              ++i                  ,++CurrIndexEntry)
            //------- We're currently in a node, so recurse down... ----
            Success = nEnumeratePage(hKWBTree, PageCount, CurrentLevel-1, PageSize, CurrIndexEntry->PageNo, fnKWEnumProc, StartKeyWord, lParam);
       }

       else

       //--------- Enumerate Page contents --------------------------
       for (Success=TRUE, next=(LPDWORD)(page+sizeof(BTREENODEHEADER)), i=0;
            Success                                        &&i<NodeHeader.NEntries;
                          next+=2                          , i++)
         Success=fnKWEnumProc(next[0], next[1], lParam);
*/
     }
     else
       Success=FALSE;

     //----------- Free Page data except Header ---------------------
     LocalFree((HLOCAL) page);
  } else
    Success = FALSE;

  return Success;
}
